@@ -5,9 +5,12 @@ use regex::Regex;
 use std::cmp::min;
 use std::fmt;
 use std::fs::File;
+use std::io::Read;
 use std::num::NonZeroUsize;
+use std::path::Path;
 use unicode_segmentation::UnicodeSegmentation;
 
+
 /* Byte formatting stuff lifted from hexyl */
 pub enum ByteCategory {
     Null,
@@ -70,6 +73,20 @@ pub const COLOR_ASCII_OTHER: Color = Color::Purple;
 pub const COLOR_NONASCII: Color = Color::Yellow;
 
 
+/// TODO Per https://doc.rust-lang.org/std/io/enum.ErrorKind.html
+/// IsADirectory isn't available yet.  Check against it when it is.
+pub fn is_a_regular_file(filename: &str) -> bool {
+    let path = Path::new(filename);
+    path.is_file()
+}
+
+
+pub fn path_exists(filename: &str) -> bool {
+    let path = Path::new(filename);
+    path.exists()
+}
+
+
 pub fn num_bytes_or_die(open_file: &Option<std::fs::File>) -> Result<usize, i32> {
     if open_file.is_none() {
         return Ok(0);
@@ -88,6 +105,63 @@ pub fn num_bytes_or_die(open_file: &Option<std::fs::File>) -> Result<usize, i32>
 }
 
 
+pub fn all_bytes_from_filename(filename: &str) -> Result<Vec<u8>, String> {
+    let file = match filehandle(filename) {
+        Ok(Some(filehandle)) => {
+            Some(filehandle)
+        },
+        Ok(None) => None,
+        Err(error) => {
+            return Err(format!("Problem opening '{}' ({:?})", filename, error));
+        }
+    };
+
+    let original_num_bytes = match num_bytes_or_die(&file) {
+        Ok(num_bytes) => {
+            num_bytes
+        },
+        Err(errcode) => {
+            return Err(format!("Couldn't read all of {}.  Errorcode {}", filename, errcode));
+        }
+    };
+
+    /* Read all bytes into memory just like real ed */
+    // TODO A real hex editor needs to buffer
+    let mut all_bytes = Vec::new();
+    if file.is_some() {
+        match file.unwrap().read_to_end(&mut all_bytes) {
+            Err(_) => {
+                if path_exists(filename) {
+                    if !is_a_regular_file(filename) {
+                        Err(format!("{} isn't a regular file.", filename))
+                    }
+                    else {
+                        Err(format!("Couldn't read {}", filename))
+                    }
+                }
+                else {
+
+                    /* Empty */
+                    Ok(all_bytes)
+                }
+            },
+            Ok(num_bytes_read) => {
+                if num_bytes_read != original_num_bytes {
+                    Err(format!("Only read {} of {} bytes of {}",
+                            num_bytes_read, original_num_bytes, filename))
+                }
+                else {
+                    Ok(all_bytes)
+                }
+            }
+        }
+    }
+    else {
+        Ok(all_bytes)
+    }
+}
+
+
 pub struct State {
     pub radix: u32,
     pub show_byte_numbers: bool,
@@ -97,6 +171,7 @@ pub struct State {
 
     pub show_prompt: bool,
     pub color: bool,
+    pub readonly: bool,
 
     /* Current byte number, 0 to (len - 1) */
     pub index: usize,
@@ -199,6 +274,16 @@ impl State {
     }
 
 
+    pub fn bytes_from_current_row(&self) -> &[u8] {
+        self.bytes_from(self.index)
+    }
+
+
+    pub fn current_row_string(&self) -> String {
+        self.bytes_line(self.bytes_from_current_row(), 0, false)
+    }
+
+
     pub fn bytes_line(&self, bytes:&[u8], line_number:usize,
             underline:bool) -> String {
         let join_char = if underline {
@@ -224,6 +309,30 @@ impl State {
     }
 
 
+    /// Return the range for the bytes in the current row
+    pub fn range_from_current_row(&self) ->
+            Result<(usize, usize), String> {
+        let last_index = self.last_byte_of_row_index();
+        if last_index.is_err() {
+            return Err(format!("{:?}", last_index));
+        }
+        Ok((self.index, last_index.unwrap()))
+    }
+
+
+    /// In current row, what's the index of the last byte
+    /// to show?
+    pub fn last_byte_of_row_index(&self) -> Result<usize, String> {
+        let max = self.max_index();
+        if max.is_err() {
+            return Err(format!("{:?}", max));
+        }
+        let max = max.unwrap();
+
+        Ok(min(max, self.index + usize::from(self.width) - 1))
+    }
+
+
     /// returns index of the byte in the 0-th column of the main row printed
     /// (not the context rows)
     pub fn print_bytes(&self) -> Option<usize> {
@@ -244,8 +353,11 @@ impl State {
         // [e..f]  after_context bytes
         let a = self.index.saturating_sub(self.before_context * usize::from(self.width));
         let b = self.index.saturating_sub(1);
-        let c = self.index;
-        let d = min(max, self.index + usize::from(self.width) - 1);
+        let cd = self.range_from_current_row();
+        if cd.is_err() {
+            return None;
+        }
+        let (c, d) = cd.unwrap();
         let e = d + 1;
         let f = min(max, self.index + (self.after_context + 1) * usize::from(self.width) - 1);
 
@@ -262,32 +374,40 @@ impl State {
         return Some(min(max, self.index + usize::from(self.width)));
     }
 
-    /// returns index of the byte in the 0-th column of the last row printed
-    /// This is more primitive than `print_bytes`.  It just prints the bytes
-    /// from the range.
-    pub fn print_bytes_sans_context(&self, range:(usize, usize),
-            underline:bool) -> Option<usize> {
+
+    pub fn bytes_in_range(&self, range:(usize, usize)) -> Result<&[u8], String> {
         if self.empty() {
-            return None;
+            return Ok(&[]);
         }
 
         let max = self.max_index();
         if max.is_err() {
-            println!("? ({:?})", max);
-            return None;
+            return Err(format!("? ({:?})", max));
         }
         let max = max.unwrap();
 
         let from = range.0;
         let to = min(max, range.1);
         if bad_range(&self.all_bytes, (from, to)) {
-            println!("? (Bad range: ({}, {}))", range.0, range.1);
-            return None;
+            return Err(format!("(Bad range: ({}, {}))", range.0, range.1));
         }
 
-        let bytes = &self.all_bytes[from..=to];
-        let max_bytes_line_num = max_bytes_line(bytes, self.width);
-        let addresses = self.addresses(from);
+        Ok(&self.all_bytes[from..=to])
+    }
+
+    /// returns index of the byte in the 0-th column of the last row printed
+    /// This is more primitive than `print_bytes`.  It just prints the bytes
+    /// from the range.
+    pub fn print_bytes_sans_context(&self, range:(usize, usize),
+            underline:bool) -> Option<usize> {
+        let bytes = self.bytes_in_range((range.0, range.1));
+        if bytes.is_err() {
+            return None;
+        }
+        let bytes = bytes.unwrap();
+
+        let max_bytes_line_num = max_bytes_line_num(bytes, self.width);
+        let addresses = self.addresses(range.0);
         if addresses.len() == 0 {
             return None;
         }
@@ -339,7 +459,8 @@ impl State {
 impl fmt::Display for State {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let mut to_write:String;
-        to_write = format!("At byte {} of {}\n", lino(&self),
+        to_write = format!("Filename: {}\n", self.filename);
+        to_write += &format!("At byte {} of {}\n", lino(&self),
                 hex_unless_dec_with_radix(self.all_bytes.len(), self.radix));
         if self.show_byte_numbers {
             to_write += &format!("Printing byte numbers in {}\n",
@@ -364,10 +485,14 @@ impl fmt::Display for State {
         };
 
         if self.unsaved_changes {
-            to_write += &format!("Unwritten changes")
+            to_write += &format!("Unwritten changes\n")
         }
         else {
-            to_write += &format!("No unwritten changes")
+            to_write += &format!("No unwritten changes\n")
+        }
+
+        if self.readonly {
+            to_write += &format!("In read-only mode\n")
         }
 
         write!(f, "{}", to_write)
@@ -480,7 +605,10 @@ pub fn move_to(state:&mut State, index:usize) -> Result<usize, String> {
 }
 
 
-pub fn max_bytes_line(bytes:&[u8], width:NonZeroUsize) -> usize {
+/// When printing [1, 2, 3, 4, 5, 6, 7] on a width of 3, the
+/// maximum 0-up line number is 3.  This function returns that
+/// calculation.  See test_max_bytes_line for examples
+pub fn max_bytes_line_num(bytes:&[u8], width:NonZeroUsize) -> usize {
     if bytes.len() == 0 {
         0
     }
@@ -510,6 +638,7 @@ mod tests {
             show_byte_numbers: true,
             show_chars: true,
             unsaved_changes: true,
+            readonly: false,
             last_search: None,
             filename: "filename".to_owned(),
             show_prompt: true,
@@ -519,6 +648,8 @@ mod tests {
             width: NonZeroUsize::new(0x10).unwrap(),
             all_bytes: Vec::from(hex_twelve),
             n_padding: "   ".to_owned(),
+            after_context: 0,
+            before_context: 0,
         };
 
         assert_eq!(state.bytes_from(0), &hex_twelve[0x00..=0x0f]);
@@ -549,11 +680,14 @@ mod tests {
             show_byte_numbers: true,
             show_chars: true,
             unsaved_changes: true,
+            readonly: false,
             last_search: None,
             filename: "filename".to_owned(),
             show_prompt: true,
             color: true,
             index: 0,
+            after_context: 0,
+            before_context: 0,
             width: NonZeroUsize::new(0x10).unwrap(),
             all_bytes: vec![
                 0x00, 0x01, 0x02,
@@ -588,46 +722,46 @@ mod tests {
         let _7 = NonZeroUsize::new(7).unwrap();
         let _8 = NonZeroUsize::new(8).unwrap();
         let bytes = vec![1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13];
-        assert_eq!(max_bytes_line(&bytes, _1), 12);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 12);
         let bytes = vec![8, 6, 7, 5, 3, 0, 9,];
-        assert_eq!(max_bytes_line(&bytes, _1), 6);
-        assert_eq!(max_bytes_line(&bytes, _2), 3);
-        assert_eq!(max_bytes_line(&bytes, _3), 2);
-        assert_eq!(max_bytes_line(&bytes, _4), 1);
-        assert_eq!(max_bytes_line(&bytes, _5), 1);
-        assert_eq!(max_bytes_line(&bytes, _6), 1);
-        assert_eq!(max_bytes_line(&bytes, _7), 0);
-        assert_eq!(max_bytes_line(&bytes, _8), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 6);
+        assert_eq!(max_bytes_line_num(&bytes, _2), 3);
+        assert_eq!(max_bytes_line_num(&bytes, _3), 2);
+        assert_eq!(max_bytes_line_num(&bytes, _4), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _5), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _6), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _7), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _8), 0);
         let bytes = vec![8, 6, 7, 5, 3, 0,];
-        assert_eq!(max_bytes_line(&bytes, _1), 5);
-        assert_eq!(max_bytes_line(&bytes, _2), 2);
-        assert_eq!(max_bytes_line(&bytes, _3), 1);
-        assert_eq!(max_bytes_line(&bytes, _4), 1);
-        assert_eq!(max_bytes_line(&bytes, _5), 1);
-        assert_eq!(max_bytes_line(&bytes, _6), 0);
-        assert_eq!(max_bytes_line(&bytes, _7), 0);
-        assert_eq!(max_bytes_line(&bytes, _8), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 5);
+        assert_eq!(max_bytes_line_num(&bytes, _2), 2);
+        assert_eq!(max_bytes_line_num(&bytes, _3), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _4), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _5), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _6), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _7), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _8), 0);
         let bytes = vec![8, 6, 7,];
-        assert_eq!(max_bytes_line(&bytes, _1), 2);
-        assert_eq!(max_bytes_line(&bytes, _2), 1);
-        assert_eq!(max_bytes_line(&bytes, _3), 0);
-        assert_eq!(max_bytes_line(&bytes, _4), 0);
-        assert_eq!(max_bytes_line(&bytes, _5), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 2);
+        assert_eq!(max_bytes_line_num(&bytes, _2), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _3), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _4), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _5), 0);
         let bytes = vec![8, 6,];
-        assert_eq!(max_bytes_line(&bytes, _1), 1);
-        assert_eq!(max_bytes_line(&bytes, _2), 0);
-        assert_eq!(max_bytes_line(&bytes, _3), 0);
-        assert_eq!(max_bytes_line(&bytes, _4), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 1);
+        assert_eq!(max_bytes_line_num(&bytes, _2), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _3), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _4), 0);
         let bytes = vec![8,];
-        assert_eq!(max_bytes_line(&bytes, _1), 0);
-        assert_eq!(max_bytes_line(&bytes, _2), 0);
-        assert_eq!(max_bytes_line(&bytes, _3), 0);
-        assert_eq!(max_bytes_line(&bytes, _4), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _2), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _3), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _4), 0);
         let bytes = vec![];
-        assert_eq!(max_bytes_line(&bytes, _1), 0);
-        assert_eq!(max_bytes_line(&bytes, _2), 0);
-        assert_eq!(max_bytes_line(&bytes, _3), 0);
-        assert_eq!(max_bytes_line(&bytes, _4), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _1), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _2), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _3), 0);
+        assert_eq!(max_bytes_line_num(&bytes, _4), 0);
     }
 
 
@@ -814,7 +948,7 @@ fn colored_chared_byte(byte:u8, underline: bool) -> String {
 }
 
 
-fn chared_byte(byte:u8) -> char {
+pub fn chared_byte(byte:u8) -> char {
     Byte(byte).as_char()
 }
 
@@ -828,14 +962,11 @@ pub fn num_graphemes(unicode_string: &str) -> usize {
 }
 
 
-/// Print version if compiled from Cargo
-pub fn print_cargo_version() {
+pub fn cargo_version() -> Result<String, String> {
     if let Some(version) = option_env!("CARGO_PKG_VERSION") {
-        println!("{}", version);
-    }
-    else {
-        println!("Version unknown (not compiled with cargo)");
+        return Ok(String::from(version));
     }
+    return Err("Version unknown (not compiled with cargo)".to_string());
 }
 
 
